/*
 *  Copyright 2001 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */


#include "fsmwrite.h"

using std::ostream;
using std::endl;

/* Write out a key of a transition. Considers the signedness of the key. */
void dumpFsmKey( ostream &out, Fsm *fsm, int onChar )
{
	if ( fsm->isAlphSigned )
		out << onChar;
	else
		out << (unsigned int) onChar;
}

/* Dump the target state, priority, and functions of a transition. */
void dumpTransData( ostream &out, Trans *trans )
{
	out << trans->toState->alg.stateNum << " [" << trans->priority << "](";

	/* Print the functions, adding in commas in between. */
	Trans::TransFuncEl *tfel = trans->transFuncTable.data;
	int nfunc = trans->transFuncTable.length();
	for ( int i = 0; i < nfunc-1; i++, tfel++ )
		out << tfel->value << ",";

	/* Print the last function without the comma. */
	if ( nfunc > 0 )
		out << tfel->value;
		
	out << ")";
}

/* Dump a transition. Trans should not be null. */
void dumpTrans( ostream &out, Fsm *fsm, int onChar, Trans *trans )
{
	out << " ";
	dumpFsmKey( out, fsm, onChar );
	out << " ";
	if ( trans != NULL )
		dumpTransData( out, trans );
	else
		out << "xx []()";
	out << " |";
}

/* Dump a range transition. */
void dumpRangePair( ostream &out, Fsm *fsm, int onChar1, int onChar2, Trans *trans )
{
	out << " ";
	dumpFsmKey( out, fsm, onChar1 );
	out << "..";
	dumpFsmKey( out, fsm, onChar2 );
	out << " ";
	if ( trans != NULL )
		dumpTransData( out, trans );
	else
		out << "xx []()";
	out << " |";
}

/* Dump a default transition. */
void dumpDefTrans( ostream &out, Trans *trans )
{
	out << " df ";
	dumpTransData( out, trans );
	out << " |";
}

/* Walk a list of transitions and dump the individual transitions. */
void dumpTransList( ostream &out, Fsm *fsm, State::TransList *list )
{
	State::TransEl *tel = list->data;
	int ntel = list->length;
	for ( int i = 0; i < ntel; i++, tel++ )
		dumpTrans( out, fsm, tel->key.intVal(), tel->value );
}

/* Walk a list of range transitions and dump the individual transitions. */
void dumpTransRange( ostream &out, Fsm *fsm, State::TransList *list )
{
	State::TransEl *tel = list->data;
	for ( int i = 0; i < list->length; i+=2, tel+=2 )
		dumpRangePair( out, fsm, tel[0].key.intVal(), tel[1].key.intVal(), tel[0].value );
}


/* Write out a state. */
void dumpState( ostream &out, Fsm *fsm, State *state )
{
	out << state->alg.stateNum << " [";
	if ( state->isOutPriorSet )
		out << state->outPriority;
	out << "](";

	/* Print the functions, adding in commas in between. */
	Trans::TransFuncEl *tfel = state->outTransFuncTable.data;
	int nfunc = state->outTransFuncTable.length();
	for ( int i = 0; i < nfunc-1; i++, tfel++ )
		out << tfel->value << ",";

	/* Print the last function without the comma. */
	if ( nfunc > 0 )
		out << tfel->value;

	out << ")\n->";

	/* If there is a default trans, dump it. */
	if ( state->outDefault != NULL )
		dumpDefTrans( out, state->outDefault );

	/* Dump the range transitions. */
	dumpTransRange( out, fsm, &state->outRange );

	/* Dump the single transitions. */
	dumpTransList( out, fsm, &state->outList );

	out << endl;
}

/* Write out a graph. */
void dumpGraph( ostream &out, Fsm *fsm )
{
	/* Set the state numbers so we print something nice! */
	fsm->setStateNumbers();

	/* Some header info. */
	out << "StartState: " << fsm->findEntry(0)->alg.stateNum << " ";
	out << "Final States:";

	/* Dump Final states. */
	State **st = fsm->finStateSet.data;
	int nst = fsm->finStateSet.length;
	for ( int i = 0; i < nst; i++, st++ )
		out << " " << (*st)->alg.stateNum << " |";
	out << endl;

	/* Walk the list of states, printing. */
	State *state = fsm->stateList.head;
	while (state) {
		dumpState( out, fsm, state );
		state = state->next;
	}
}
