/*
 *  Copyright 2001, 2002 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#include <iostream>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include "ragel.h"
#include "parsetree.h"
#include "fsmmachine.h"
#include "fsmbuild.h"

/* Code generators. */
#include "tabcodegen.h"
#include "ftabcodegen.h"
#include "gotocodegen.h"
#include "fgotocodegen.h"
#include "ipgcodegen.h"

using std::cerr;
using std::cout;
using std::endl;

/* The minimum chars are the absolute value of the real minimum because the
 * sign is stored separately in integers read in so they are compared in the
 * positive. Each is casted to an unsigned because the data part of the number
 * is in unsigned int size. 
 *
 * Not terribly portable...
 */
#define RL_MIN_CHAR    ((int)((char)0x80))
#define RL_MAX_CHAR    ((int)((char)0x7f))
#define RL_MIN_UCHAR   ((int)((unsigned char)0x0))
#define RL_MAX_UCHAR   ((int)((unsigned char)0xff))
#define RL_MIN_SHORT   ((int)((short)0x8000))
#define RL_MAX_SHORT   ((int)((short)0x7fff))
#define RL_MIN_USHORT  ((int)((unsigned short)0x0000))
#define RL_MAX_USHORT  ((int)((unsigned short)0xffff))
#define RL_MIN_INT     ((int)((int)0x80000000))
#define RL_MAX_INT     ((int)((int)0x7fffffff))
#define RL_MIN_UINT    ((int)((unsigned int)0x00000000))
#define RL_MAX_UINT    ((int)((unsigned int)0xffffffff))

int curFuncOrdering = 0;

/* Perform minimization after an operation according 
 * to the command line args. */
void afterOpMinimize( Fsm *fsm )
{
	/* Switch on the prefered minimization algorithm. */
	if ( minimizeEveryOp ) {
		if ( minimizeLevel != MinimizeNone ) {
			/* First clean up the graph. Fsm operations may leave these lying
			 * around.  There should be no dead end states however. The
			 * subtract intersection operators are the only places where they
			 * may be created and those operators clean them up. */
			fsm->removeUnreachableStates();
		}

		switch ( minimizeLevel ) {
			case MinimizeNone:
				break;
			case MinimizeApprox:
				fsm->minimizeApproximate();
				break;
			case MinimizePartition1:
				fsm->minimizePartition1();
				break;
			case MinimizePartition2:
				fsm->minimizePartition2();
				break;
			case MinimizeStable:
				fsm->minimizeStable();
				break;
		}
	}
}

/* Count the transitions in the fsm by walking the state list. */
int countTransitions( Fsm *fsm )
{
	int numTrans = 0;
	State *state = fsm->stateList.head;
	while ( state != NULL ) {
		numTrans += state->outList.length;
		numTrans += state->outRange.length / 2;
		if ( state->outDefault != 0 )
			numTrans += 1;
		state = state->next;
	}
	return numTrans;
}

/* Count the indicies in a fsm machine by walking array of states. */
int countIndicies( FsmMachine &machine )
{
	int sum = 0;
	for ( int state = 0; state < machine.numStates; state++ )
		sum += machine.allStates[state].numIndex;
	return sum;
}

int makeFsmKeyHex( char *str, int lineNum, ParseData *pd )
{
	int retVal = 0;

	/* Convert the number to a decimal. Reset errno so we can check for
	 * overflow or underflow. */
	errno = 0;
	unsigned int ul = strtoul( str, 0, 16 );

	switch ( pd->alphType ) {
	case AT_Char: {
		/* Check against upper bound. */
		if ( ul > RL_MAX_UCHAR ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows char alphabet" << endl;
			exit(1);
		}

		/* Cast to a char to get the proper sign, then extend to an integer. */
		retVal = (char)ul;
		break;
	}

	case AT_UnsignedChar: {
		/* Check against upper bound. */
		if ( ul > RL_MAX_UCHAR ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows unsigned char alphabet" << endl;
			exit(1);
		}

		/* Store cast to unsigned char, then extend to an integer. */
		retVal = (unsigned char)ul;
		break;
	}

	case AT_Short: {
		/* Check against upper bound. */
		if ( ul > RL_MAX_USHORT ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows short alphabet" << endl;
			exit(1);
		}

		/* Cast to a short to get the proper sign, then extend to an integer. */
		retVal = (short)ul;
		break;
	}

	case AT_UnsignedShort: {
		/* Check against upper bound. */
		if ( ul > RL_MAX_USHORT ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows unsigned short alphabet" << endl;
			exit(1);
		}

		/* Store cast to unsigned short, then extend to an integer. */
		retVal = (unsigned short)ul;
		break;
	}

	case AT_Int: {
		/* Check against upper bound. */
		if ( errno == ERANGE && ul == ULONG_MAX ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows int alphabet" << endl;
			exit(1);
		}

		/* Store in integer format. */
		retVal = (int)ul;
		break;
	}

	case AT_UnsignedInt: {
		/* Check against upper bound. */
		if ( errno == ERANGE && ul == ULONG_MAX ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows unsigned int alphabet" << endl;
			exit(1);
		}

		/* Store in integer format. */
		retVal = (int)ul;
		break;
	}}
	return retVal;
}

int makeFsmKeyDec( char *str, int lineNum, ParseData *pd )
{
	int retVal = 0;

	switch ( pd->alphType ) {
	case AT_Char: {
		/* Convert the number to a decimal. If there is overflow or underflow
		 * the num will be set to LONG_MIN/LONG_MAX which are way too low and
		 * high for char anyways so no need to check. */
		retVal = strtol( str, 0, 10 );

		/* Check against lower bound. */
		if ( retVal < RL_MIN_CHAR ) {
	 		cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" underflows char alphabet" << endl;
			exit(1);
		}

		/* Check against upper bound. */
		if ( retVal > RL_MAX_CHAR ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows char alphabet" << endl;
			exit(1);
		}
		break;
	}

	case AT_UnsignedChar: {
		/* Can't allow a negative decimal number for unsigned char type. */
		if ( str[0] == '-' ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" underflows unsigned char alphabet" << endl;
			exit(1);
		}

		/* Convert to an unsigned number and bounds check it. */
		unsigned int ul = strtoul( str, 0, 10 );
		if ( ul > RL_MAX_UCHAR ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows unsigned char alphabet" << endl;

		}

		/* Store in integer format. */
		retVal = (int) ul;
		break;
	}

	case AT_Short: {
		/* Convert the number to a decimal. If there is overflow or underflow
		 * the num will be set to the high or low limits which are exceed the
		 * bounds for short so no need to check. */
		retVal = strtol( str, 0, 10 );

		/* Check against lower bound. */
		if ( retVal < RL_MIN_SHORT ) {
	 		cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" underflows short alphabet" << endl;
			exit(1);
		}

		/* Check against upper bound. */
		if ( retVal > RL_MAX_SHORT ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows short alphabet" << endl;
			exit(1);
		}
		break;
	}

	case AT_UnsignedShort: {
		/* Can't allow a negative decimal number for unsigned short type. */
		if ( str[0] == '-' ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" underflows unsigned short alphabet" << endl;
			exit(1);
		}

		/* Convert to an unsigned number and bounds check it. */
		unsigned int ul = strtoul( str, 0, 10 );
		if ( ul > RL_MAX_USHORT ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows unsigned short alphabet" << endl;

		}

		/* Store in integer format. */
		retVal = (int) ul;
		break;
	}

	case AT_Int: {
		/* Convert the number to a decimal. First reset errno so we can check
		 * for overflow or underflow. */
		errno = 0;
		retVal = strtol( str, 0, 10 );

		/* Check lower bound. */
		if ( errno == ERANGE && retVal == LONG_MIN ) {
	 		cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" underflows int alphabet" << endl;
			exit(1);
		}

		/* Check upper bound. */
		if ( errno == ERANGE && retVal == LONG_MAX ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows int alphabet" << endl;
			exit(1);
		}
		break;
	}

	case AT_UnsignedInt: {
		/* Cannot allow negative numbers for an unsigned type. */
		if ( str[0] == '-' ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" underflows unsigned int alphabet" << endl;
			exit(1);
		}

		/* Convert to an unsigned number and bounds check it. */
		errno = 0;
		unsigned int ul = strtoul( str, 0, 10 );
		if ( errno == ERANGE && ul == ULONG_MAX ) {
			cerr << pd->fileName << ":" << lineNum << ": literal " << str <<
					" overflows unsigned int alphabet" << endl;
		}

		/* Store in integer format. */
		retVal = (int) ul;
		break;
	}}
	return retVal;
}

/* Make an fsm key in int format (what the fsm graph uses) from an alphabet
 * number returned by the parser. Validates that the number doesn't overflow
 * the alphabet type. */
int makeFsmKeyNum( char *str, int lineNum, ParseData *pd )
{
	/* Switch on hex/decimal format. */
	if ( str[0] == '0' && str[1] == 'x' )
		return makeFsmKeyHex( str, lineNum, pd );
	else
		return makeFsmKeyDec( str, lineNum, pd );

}

/* Make an fsm int format (what the fsm graph uses) from a single character.
 * Performs proper conversion depending on signed/unsigned property of the
 * alphabet. */
int makeFsmKeyChar( char c, ParseData *pd )
{
	int retVal;
	if ( pd->isAlphSigned() ) {
		/* Copy from a char type. */
		retVal = c;
	}
	else {
		/* Copy from an unsigned byte type. */
		retVal = (unsigned char)c;
	}
	return retVal;
}

/* Make an fsm key array in int format (what the fsm graph uses) from a string
 * of characters. Performs proper conversion depending on signed/unsigned
 * property of the alphabet. */
void makeFsmKeyArray( int *result, char *data, int len, ParseData *pd )
{
	if ( pd->isAlphSigned() ) {
		/* Copy from a char star type. */
		char *src = data;
		for ( int i = 0; i < len; i++ )
			result[i] = src[i];
	}
	else {
		/* Copy from an unsigned byte ptr type. */
		unsigned char *src = (unsigned char*) data;
		for ( int i = 0; i < len; i++ )
			result[i] = src[i];
	}
}

/* Make a builtin type. Depends on the signed nature of the alphabet type. */
Fsm *makeBuiltin( BuiltinType builtin, ParseData *pd )
{
	/* Fsm created to return. */
	Fsm *retFsm = 0;
	bool isSigned = pd->isAlphSigned();

	switch ( builtin ) {
	case BT_Any:
		/* All characters. */
		retFsm = Fsm::dotFsm( isSigned );
		break;

	case BT_Ascii:
		/* Ascii characters 0 to 127. */
		retFsm = Fsm::rangeFsm( 0, 127, isSigned );
		break;

	case BT_Extend:
		/* Ascii extended characters. This is the full byte range. Dependent
		 * on signed, vs no signed. If the alphabet is one byte then just use
		 * dot fsm. */
		if ( pd->alphType == AT_Char || pd->alphType == AT_UnsignedChar )
			retFsm = Fsm::dotFsm( isSigned );
		else {
			retFsm = isSigned ? 
					Fsm::rangeFsm( -128, 127, true ) : 
					Fsm::rangeFsm( 0, 255, false );
		}
		break;

	case BT_Alpha:
		/* Alpha [A-Za-z]. */
		retFsm = Fsm::rangeFsm( 'A', 'Z', isSigned );
		retFsm->unionOp( Fsm::rangeFsm( 'a', 'z', isSigned ) );
		retFsm->minimizePartition2();
		break;

	case BT_Digit:
		/* Digits [0-9]. */
		retFsm = Fsm::rangeFsm( '0', '9', isSigned );
		break;

	case BT_Alnum:
		/* Alpha numerics [0-9A-Za-z]. */
		retFsm = Fsm::rangeFsm( '0', '9', isSigned );
		retFsm->unionOp( Fsm::rangeFsm( 'A', 'Z', isSigned ) );
		retFsm->unionOp( Fsm::rangeFsm( 'a', 'z', isSigned ) );
		retFsm->minimizePartition2();
		break;

	case BT_Lower:
		/* Lower case characters. */
		retFsm = Fsm::rangeFsm( 'a', 'z', isSigned );
		break;

	case BT_Upper:
		/* Upper case characters. */
		retFsm = Fsm::rangeFsm( 'A', 'Z', isSigned );
		break;

	case BT_Cntrl:
		/* Control characters. */
		retFsm = Fsm::rangeFsm( 0, 31, isSigned );
		retFsm->unionOp( Fsm::concatFsm( 127, isSigned ) );
		retFsm->minimizePartition2();
		break;

	case BT_Graph:
		/* Graphical ascii characters [!-~]. */
		retFsm = Fsm::rangeFsm( '!', '~', isSigned );
		break;

	case BT_Print:
		/* Printable characters. Same as graph except includes space. */
		retFsm = Fsm::rangeFsm( ' ', '~', isSigned );
		break;

	case BT_Punct:
		/* Punctuation. */
		retFsm = Fsm::rangeFsm( '!', '/', isSigned );
		retFsm->unionOp( Fsm::rangeFsm( ':', '@', isSigned ) );
		retFsm->unionOp( Fsm::rangeFsm( '[', '`', isSigned ) );
		retFsm->unionOp( Fsm::rangeFsm( '{', '~', isSigned ) );
		retFsm->minimizePartition2();
		break;

	case BT_Space:
		/* Whitespace: [\t\v\f\n\r ]. */
		retFsm = Fsm::rangeFsm( '\t', '\r', isSigned );
		retFsm->unionOp( Fsm::concatFsm( ' ', isSigned ) );
		retFsm->minimizePartition2();
		break;

	case BT_Xdigit:
		/* Hex digits [0-9A-Fa-f]. */
		retFsm = Fsm::rangeFsm( '0', '9', isSigned );
		retFsm->unionOp( Fsm::rangeFsm( 'A', 'F', isSigned ) );
		retFsm->unionOp( Fsm::rangeFsm( 'a', 'f', isSigned ) );
		retFsm->minimizePartition2();
		break;
	}

	return retFsm;
}


/*
 * ParseData
 */

/* Initialize the structure that will collect info during the parse of a
 * machine. */
ParseData::ParseData( const String &fileName )
:	
	nextFuncNum(0), 
	machineGiven(false), 
	alphType(AT_Char),
	fileName(fileName)
{
	/* Initialize the dictionary of graphs. This is our symbol table. The
	 * initialization needs to be done on construction which happens at the
	 * beginning of a machine spec so any assignment operators can reference
	 * the builtins. */
	initGraphDict( );
}

/* Clean up the data collected during a parse. */
ParseData::~ParseData()
{
	/* Delete all the nodes in the function list. Will cause all the
	 * string data that represents the functions to be deallocated. */
	funcList.empty();
	dataList.empty();
	initCodeList.empty();
}

/* Is the alphabet type a signed type? */
bool ParseData::isAlphSigned()
{
	return alphType == AT_Char || alphType == AT_Short || alphType == AT_Int;
}

/* Initialize the graph dict with builtin types. */
void ParseData::initGraphDict( )
{
	/* All characters. */
	graphDict.insert( "any", new ExpressionNode( BT_Any ) );
	graphDict.insert( "ascii", new ExpressionNode( BT_Ascii ) );
	graphDict.insert( "extend", new ExpressionNode( BT_Extend ) );
	graphDict.insert( "alpha", new ExpressionNode( BT_Alpha ) );
	graphDict.insert( "digit", new ExpressionNode( BT_Digit ) );
	graphDict.insert( "alnum", new ExpressionNode( BT_Alnum ) );
	graphDict.insert( "lower", new ExpressionNode( BT_Lower ) );
	graphDict.insert( "upper", new ExpressionNode( BT_Upper ) );
	graphDict.insert( "cntrl", new ExpressionNode( BT_Cntrl ) );
	graphDict.insert( "graph", new ExpressionNode( BT_Graph ) );
	graphDict.insert( "print", new ExpressionNode( BT_Print ) );
	graphDict.insert( "punct", new ExpressionNode( BT_Punct ) );
	graphDict.insert( "space", new ExpressionNode( BT_Space ) );
	graphDict.insert( "xdigit", new ExpressionNode( BT_Xdigit ) );
}

/* Set the alphabet type. If type types are not valid returns false. */
bool ParseData::setAlphType( String s1, String s2 )
{
	bool valid = false;
	if ( strcmp( s1, "unsigned" ) == 0 ) {
		if ( strcmp( s2, "char" ) == 0 ) {
			alphType = AT_UnsignedChar;
			valid = true;
		}
		else if ( strcmp( s2, "short" ) == 0 ) {
			alphType = AT_UnsignedShort;
			valid = true;
		}
		else if ( strcmp( s2, "int" ) == 0 ) {
			alphType = AT_UnsignedInt;
			valid = true;
		}
	}
	return valid;
}

/* Set the alphabet type. If type types are not valid returns false. */
bool ParseData::setAlphType( String s1 )
{
	bool valid = false;
	if ( strcmp( s1, "char" ) == 0 ) {
		alphType = AT_Char;
		valid = true;
	}
	else if ( strcmp( s1, "short" ) == 0 ) {
		alphType = AT_Short;
		valid = true;
	}
	else if ( strcmp( s1, "int" ) == 0 ) {
		alphType = AT_Int;
		valid = true;
	}
	return valid;
}

void ParseData::dumpWithHeader( FsmAp *graph, const char *head )
{
	cout << endl << fsmName << head << endl;
	for ( int count = fsmName.getLen() + strlen(head); count > 0; count-- )
		cout << '=';
	cout << endl;
	dumpGraph( cout, graph );
	cout << endl;
}

void ParseData::setLowerUpperRange( )
{
	if ( lowerNum.getLen() != 0 ) {
		/* If ranges are given then interpret the alphabet type. */
		lowerKey = makeFsmKeyNum( lowerNum, rangeLineNum, this );
		upperKey = makeFsmKeyNum( upperNum, rangeLineNum, this );
	}
	else {
		/* Use default ranges for alphabet type. */
		switch ( alphType ) {
		case AT_Char:
			lowerKey = RL_MIN_CHAR;
			upperKey = RL_MAX_CHAR;
			break;
		case AT_UnsignedChar:
			lowerKey = RL_MIN_UCHAR;
			upperKey = RL_MAX_UCHAR;
			break;
		case AT_Short:
			lowerKey = RL_MIN_SHORT;
			upperKey = RL_MAX_SHORT;
			break;
		case AT_UnsignedShort:
			lowerKey = RL_MIN_USHORT;
			upperKey = RL_MAX_USHORT;
			break;
		case AT_Int:
			lowerKey = RL_MIN_INT;
			upperKey = RL_MAX_INT;
			break;
		case AT_UnsignedInt:
			lowerKey = RL_MIN_UINT;
			upperKey = RL_MAX_UINT;
			break;
		}
	}
}


/* Generate the code for an fsm. Assumes parseData is set up properly. Called
 * by parser code. */
void ParseData::generateFsm( )
{
	/* The machine that will take the compressed fsm. */
	FsmMachine fsmMachine;

	/* Make a code generator that will output the header/code. */
	FsmCodeGen *codeGen = 0;
	switch ( codeGeneration ) {
	case C_Code:
		switch ( codeStyle ) {
		case GenTables:
			codeGen = new CTabCodeGen( fsmName, &fsmMachine, this, *outputStream );
			break;
		case GenFTables:
			codeGen = new CFTabCodeGen( fsmName, &fsmMachine, this, *outputStream );
			break;
		case GenGoto:
			codeGen = new CGotoCodeGen( fsmName, &fsmMachine, this, *outputStream );
			break;
		case GenFGoto:
			codeGen = new CFGotoCodeGen( fsmName, &fsmMachine, this, *outputStream );
			break;
		case GenIpGoto:
			codeGen = new CIpGotoCodeGen( fsmName, &fsmMachine, this, *outputStream );
			break;
		}
		break;
	case CC_Code:
		switch ( codeStyle ) {
		case GenTables:
			codeGen = new CCTabCodeGen( fsmName, &fsmMachine, this, *outputStream );
			break;
		case GenFTables:
			codeGen = new CCFTabCodeGen( fsmName, &fsmMachine, this, *outputStream );
			break;
		case GenGoto:
			codeGen = new CCGotoCodeGen( fsmName, &fsmMachine, this, *outputStream );
			break;
		case GenFGoto:
			codeGen = new CCFGotoCodeGen( fsmName, &fsmMachine, this, *outputStream );
			break;
		case GenIpGoto:
			codeGen = new CCIpGotoCodeGen( fsmName, &fsmMachine, this, *outputStream );
			break;
		}
		break;
	}

	/* If the struct element was used then output the header portion. */
	if ( dataList.length > 0 )
		codeGen->writeOutHeader( );

	/* If the graph dict is given, then output the code portion. */
	if ( machineGiven ) {
		/* Get the main line from the graph dict. */
		GraphDictNode *gdNode = graphDict.find("main");
		if ( gdNode == NULL ) {
			cerr << fileName << ":" << fsmName << ": " << MSG_NO_MAIN_GRAPH << endl;
			exit(1);
		}

		/* Interpret the alphabet type. */
		setLowerUpperRange( );

		/* Build the graph from a walk of the parse tree. */
		Fsm *graph = gdNode->value->walk( this );

		/* Do final removal of unreachable states. There should be no dead end
		 * states however. The subtract and intersection operators are the
		 * only places where they may be created and those operators clean
		 * them up. */
		graph->removeUnreachableStates();

		/* Dump of graph before minmization (to see priorities). */
		if ( dumpFinalPrior )
			dumpWithHeader( graph, " Before Minimization" );

		/* If minimizing we can optimize the graph somewhat. */
		if ( minimizeLevel != MinimizeNone ) {
			/* No more fsm operations are to be done. Function ordering numbers are
			 * no longer of use and will just hinder minimization. Clear them. */
			graph->nullFunctionKeys();

			/* Transition priorities are no longer of use. We can clear them
			 * because they will just hinder minimization as well. Clear them. */
			graph->allTransPrior( 0 );			
			graph->clearLeaveFsmPrior();
		}

		/* Minimize here even if we minimized at every op. Now that function
		 * keys have been cleared we may get a more minimal fsm. */
		switch ( minimizeLevel ) {
			case MinimizeNone:
				break;
			case MinimizeApprox:
				graph->minimizeApproximate();
				break;
			case MinimizeStable:
				graph->minimizeStable();
				break;
			case MinimizePartition1:
				graph->minimizePartition1();
				break;
			case MinimizePartition2:
				graph->minimizePartition2();
				break;
		}

		/* Dump of graph before minmization (to see priorities). */
		if ( dumpFinal )
			dumpWithHeader( graph, " After Minimization" );

		if ( printStats ) {
			cout << "graph states:      " << graph->stateList.length << endl;
			cout << "graph transitions: " << countTransitions( graph ) << endl;
		}

		/* Build the machine from the graph and output it. */
		FsmBuild fsmBuild( *graph, fsmMachine );
		fsmBuild.build();

		if ( printStats ) {
			cout << "machine states:    " << fsmMachine.numStates << endl;
			cout << "machine trans:     " << fsmMachine.numTrans << endl;
			cout << "machine indicies:  " << fsmMachine.numIndicies() << endl;
			cout << "machine functions: " << fsmMachine.numTransFuncIndex << endl;
			cout << "function array:    " << fsmMachine.numTransFuncs << endl;
		}

		codeGen->writeOutCode( );
	}
}


/* Clean up after an expression node. */
ExpressionNode::~ExpressionNode()
{
	switch ( type ) {
		case Or: case Intersect: case Subtract:
			delete expression;
			delete term;
			break;
		case Term:
			delete term;
			break;
		case Builtin:
			break;
	}
}


/* Evaluate and expression node. */
Fsm *ExpressionNode::walk( ParseData *pd )
{
	Fsm *rtnVal = 0;
	switch ( type ) {
		case Or: {
			/* Evaluate the expression. */
			rtnVal = expression->walk( pd );
			/* Evaluate the term. */
			Fsm *rhs = term->walk( pd );
			/* Perform union. */
			rtnVal->unionOp( rhs );
			/* Minimization. */
			afterOpMinimize( rtnVal );
			break;
		}
		case Intersect: {
			/* Evaluate the expression. */
			rtnVal = expression->walk( pd );
			/* Evaluate the term. */
			Fsm *rhs = term->walk( pd );
			/* Perform intersection. */
			rtnVal->intersectOp( rhs );
			/* Minimization. */
			afterOpMinimize( rtnVal );
			break;
		}
		case Subtract: {
			/* Evaluate the expression. */
			rtnVal = expression->walk( pd );
			/* Evaluate the term. */
			Fsm *rhs = term->walk( pd );
			/* Perform subtraction. */
			rtnVal->subtractOp( rhs );
			/* Minimization. */
			afterOpMinimize( rtnVal );
			break;
		}
		case Term: {
			/* Return result of the term. */
			rtnVal = term->walk( pd );
			break;
		}
		case Builtin: {
			/* Duplicate the builtin. */
			rtnVal = makeBuiltin( builtin, pd );
			break;
		}
	}

	return rtnVal;
}

/* Clean up after a term node. */
TermNode::~TermNode()
{
	switch ( type ) {
		case Concat:
			delete term;
			delete factorWithAug;
			break;
		case FactorWithAug:
			delete factorWithAug;
			break;
	}
}

/* Evaluate a term node. */
Fsm *TermNode::walk( ParseData *pd )
{
	Fsm *rtnVal = 0;
	switch ( type ) {
		case Concat: {
			/* Evaluate the Term. */
			rtnVal = term->walk( pd );
			/* Evaluate the FactorWithRep. */
			Fsm *rhs = factorWithAug->walk( pd );
			/* Perform concatenation. */
			rtnVal->concatOp( rhs, leavingFsm );
			/* Minimization. */
			afterOpMinimize( rtnVal );
			break;
		}
		case FactorWithAug: {
			rtnVal = factorWithAug->walk( pd );
			break;
		}
	}
	return rtnVal;
}

/* Clean up after a factor with augmentation node. */
FactorWithAugNode::~FactorWithAugNode()
{
	delete factorWithRep;

	/* Walk the vector of parser funcs, deleting function names. */
}

/* Evaluate a factor with augmentation node. */
Fsm *FactorWithAugNode::walk( ParseData *pd )
{
	/* First walk the list of functions, assigning the
	 * starting function ordering. */
	ParserFunc *func = funcs.data;
	int i, nFuncs = funcs.length;
	for ( i = 0; i < nFuncs; i++, func++ ) {
		if ( func->type == start )
			func->ordering = curFuncOrdering++;
	}

	/* Evaluate the factor with repetition. */
	Fsm *rtnVal = factorWithRep->walk( pd );

	/* Now compute the remaining function call orderings. */
	func = funcs.data;
	nFuncs = funcs.length;
	for ( i = 0; i < nFuncs; i++, func++ ) {
		if ( func->type != start )
			func->ordering = curFuncOrdering++;
	}

	/* Assign functions. */
	func = funcs.data;
	nFuncs = funcs.length;
	for ( i = 0; i < nFuncs; i++, func++ )  {
		if ( func->func == FUNC_NUM_CLEAR ) {
			switch ( func->type ) {
			case start:
				/* Changing start functions affects potential minimization. */
				rtnVal->clearStartFsmFunc( );
				afterOpMinimize( rtnVal );
				break;
			case all:
				rtnVal->clearAllTransFunc( );
				break;
			case fin:
				rtnVal->clearFinFsmFunc( );
				break;
			case leave:
				rtnVal->clearLeaveFsmFunc( );
				break;
			case clearLeave:
				assert(false);
			}
		}
		else {
			switch ( func->type ) {
			case start:
				/* Start fsm funcs are a special case that may require
				 * minimization afterwards. */
				rtnVal->startFsmFunc( func->func, func->ordering );
				afterOpMinimize( rtnVal );
				break;
			case all:
				rtnVal->allTransFunc( func->func, func->ordering );
				break;
			case fin:
				rtnVal->finFsmFunc( func->func, func->ordering );
				break;
			case leave:
				rtnVal->leaveFsmFunc( func->func, func->ordering );
				break;
			case clearLeave:
				assert(false);
			}
		}
	}

	/* Assign priorities. */
	PriorityAug *aug = priorityAugs.data;
	int nAugs = priorityAugs.length;
	for ( i = 0; i < nAugs; i++, aug++ ) {
		switch ( aug->type ) {
			case start:
				rtnVal->startFsmPrior( aug->value );
				/* Start fsm priorities are a special case that may require
				 * minimization afterwards. */
				afterOpMinimize( rtnVal );
				break;
			case all:
				rtnVal->allTransPrior( aug->value );
				break;
			case fin:
				rtnVal->finFsmPrior( aug->value );
				break;
			case leave:
				rtnVal->leaveFsmPrior( aug->value );
				break;
			case clearLeave:
				rtnVal->clearLeaveFsmPrior();
				break;
		}
	}
	
	return rtnVal;
}

/* Clean up after a factor with repetition node. */
FactorWithRepNode::~FactorWithRepNode()
{
	switch ( type ) {
		case Star: case Optional: case Plus: case Negate:
			delete factorWithRep;
			break;
		case Factor:
			delete factor;
			break;
	}
}

/* Evaluate a factor with repetition node. */
Fsm *FactorWithRepNode::walk( ParseData *pd )
{
	Fsm *rtnVal = 0;
	bool isSigned = pd->isAlphSigned();

	switch ( type ) {
		case Star: {
			/* Evaluate the FactorWithRep. */
			rtnVal = factorWithRep->walk( pd );
			/* Perform the kleen star. */
			if ( leavingFsm ) {
				curFuncOrdering +=
						rtnVal->shiftStartFuncOrder( curFuncOrdering );
			}
			rtnVal->starOp(leavingFsm);
			/* Minimization. */
			afterOpMinimize(rtnVal);
			break;
		}
		case Optional: {
			/* Evaluate the FactorWithRep. */
			rtnVal = factorWithRep->walk( pd );
			/* Perform the question operator. */
			rtnVal->unionOp( Fsm::nullFsm( isSigned ) );
			/* Minimization. */
			afterOpMinimize(rtnVal);
			break;
		}
		case Plus: {
			/* Evaluate the FactorWithRep. */
			rtnVal = factorWithRep->walk( pd );
			/* Perform the plus operator. */
			Fsm *dup = new Fsm(*rtnVal);
			if ( leavingFsm ) {
				curFuncOrdering +=
						dup->shiftStartFuncOrder( curFuncOrdering );
			}
			dup->starOp(leavingFsm);
			/* Intermediate minimization. */
			afterOpMinimize(dup);
			rtnVal->concatOp(dup, leavingFsm);
			/* Minimization. */
			afterOpMinimize(rtnVal);
			break;
		}
		case Negate: {
			/* Evaluate the FactorWithRep. */
			Fsm *toNegate = factorWithRep->walk( pd );
			/* Negation is subtract from dot-star. */
			rtnVal = Fsm::dotStarFsm( isSigned );
			rtnVal->subtractOp( toNegate );
			/* Minimization. */
			afterOpMinimize(rtnVal);
			break;
		}
		case Factor: {
			/* Evaluate the Factor. Pass it up. */
			rtnVal = factor->walk( pd );
			break;
		}
	}
	return rtnVal;
}

/* Clean up after a factor node. */
FactorNode::~FactorNode()
{
	switch ( type ) {
		case LiteralFsm:
			delete literal;
			break;
		case RangeFsm:
			delete range;
			break;
		case RegularExpression:
			delete regExp;
			break;
		case LookupExpression:
			break;
		case Expression:
			delete expression;
			break;
	}
}

/* Evaluate a factor node. */
Fsm *FactorNode::walk( ParseData *pd )
{
	Fsm *rtnVal = 0;
	switch ( type ) {
		case LiteralFsm:
			rtnVal = literal->walk( pd );
			break;
		case RangeFsm:
			rtnVal = range->walk( pd );
			break;
		case RegularExpression:
			rtnVal = regExp->walk( pd );
			break;
		case LookupExpression:
		case Expression:
			rtnVal = expression->walk( pd );
			break;
	}
	return rtnVal;
}

/* Clean up a range object. Must delete the two literals. */
Range::~Range()
{
	delete lowerLit;
	delete upperLit;
}

/* Evaluate a range. Gets the lower an upper key and makes an fsm range. */
Fsm *Range::walk( ParseData *pd )
{
	int lowerKey = lowerLit->makeRangeEndPoint( pd );
	int upperKey = upperLit->makeRangeEndPoint( pd );

	/* Validate the range. */
	if ( FsmKey::gt( FsmKey(lowerKey), FsmKey(upperKey), pd->isAlphSigned() ) ) {
		cerr << pd->fileName << ":" << lowerLit->lineNum << ": lower end of"
			" range " << lowerLit->str << " is greater then upper end " << 
			upperLit->str << endl;
		exit(1);
	}
	return Fsm::rangeFsm( lowerKey, upperKey, pd->isAlphSigned() );
}

/* Return an fsm key for a literal that is a range end point. */
int Literal::makeRangeEndPoint( ParseData *pd )
{
	int retVal;
	switch ( type ) {
	case Number: {
		/* Make the fsm key from the number. */
		retVal = makeFsmKeyNum( str, lineNum, pd );
		break;
	}
	case LitString: {
		/* The string should be verified length one. */
		makeFsmKeyArray( &retVal, str, 1, pd );
		break;
	}
	case OrLitString: {
		/* Should not be used for a range. */
		assert( false );
		break;
	}}
	return retVal;
}

/* Evaluate a literal object. */
Fsm *Literal::walk( ParseData *pd )
{
	/* Fsm to return, is the alphabet signed. */
	Fsm *rtnVal = 0;
	bool isSigned = pd->isAlphSigned();

	switch ( type ) {
	case Number: {
		int fsmKey = makeFsmKeyNum( str, lineNum, pd );
		rtnVal = Fsm::concatFsm( fsmKey, isSigned );
		break;
	}
	case LitString: {
		int arr[str.getLen()];
		makeFsmKeyArray( arr, str, str.getLen(), pd );
		rtnVal = Fsm::concatFsm( arr, str.getLen(), isSigned );
		break;
	}
	case OrLitString: {
		int arr[str.getLen()];
		makeFsmKeyArray( arr, str, str.getLen(), pd );
		rtnVal = Fsm::orFsm( arr, str.getLen(), isSigned );
		break;
	}}
	return rtnVal;
}

/* Clean up after a regular expression object. */
RegExp::~RegExp()
{
	switch ( type ) {
		case RecurseItem:
			delete regExp;
			delete item;
			break;
		case Empty:
			break;
	}
}

/* Evaluate a regular expression object. */
Fsm *RegExp::walk( ParseData *pd )
{
	Fsm *rtnVal = 0;
	switch ( type ) {
		case RecurseItem: {
			/* Walk both items. */
			Fsm *fsm1 = regExp->walk( pd );
			Fsm *fsm2 = item->walk( pd );
			if ( fsm1 == 0 )
				rtnVal = fsm2;
			else {
				fsm1->concatOp( fsm2, false );
				rtnVal = fsm1;
			}
			break;
		}
		case Empty: {
			rtnVal = 0;
			break;
		}
	}
	return rtnVal;
}

/* Clean up after an item in a regular expression. */
ReItem::~ReItem()
{
	switch ( type ) {
		case Data:
		case Dot:
			break;
		case OrBlock:
		case NegOrBlock:
			delete orBlock;
			break;
	}
}

/* Evaluate a regular expression object. */
Fsm *ReItem::walk( ParseData *pd )
{
	/* The fsm to return, is the alphabet signed? */
	Fsm *rtnVal = 0;
	bool isSigned = pd->isAlphSigned();

	switch ( type ) {
		case Data: {
			/* Move the data into an integer array and make a concat fsm. */
			int arr[str.getLen()];
			makeFsmKeyArray( arr, str, str.getLen(), pd );
			rtnVal = Fsm::concatFsm( arr, str.getLen(), isSigned );

			/* If the data item is stared, it should be of length one. */
			assert( ! (str.getLen() > 1 && star ) );
			break;
		}
		case Dot: {
			rtnVal = Fsm::dotFsm( isSigned );
			break;
		}
		case OrBlock: {
			/* Get the or block and minmize it. */
			rtnVal = orBlock->walk( pd );
			rtnVal->minimizePartition2();
			break;
		}
		case NegOrBlock: {
			/* Get the or block and minimize it. */
			Fsm *fsm = orBlock->walk( pd );
			fsm->minimizePartition2();

			/* Make a dot fsm and subtract from it. */
			rtnVal =  Fsm::dotFsm( isSigned );
			rtnVal->subtractOp( fsm );
			rtnVal->minimizePartition2();
			break;
		}
	}

	/* If the item is followed by a star, then apply the star op. */
	if ( star ) {
		rtnVal->starOp(false);
		rtnVal->minimizePartition2();
	}
	return rtnVal;
}

/* Clean up after an or block of a regular expression. */
ReOrBlock::~ReOrBlock()
{
	switch ( type ) {
		case RecurseItem:
			delete orBlock;
			delete item;
			break;
		case Empty:
			break;
	}
}


/* Evaluate an or block of a regular expression. */
Fsm *ReOrBlock::walk( ParseData *pd )
{
	Fsm *rtnVal = 0;
	switch ( type ) {
		case RecurseItem: {
			/* Evaluate the two fsm. */
			Fsm *fsm1 = orBlock->walk( pd );
			Fsm *fsm2 = item->walk( pd );
			if ( fsm1 == 0 )
				rtnVal = fsm2;
			else {
				fsm1->unionOp( fsm2 );
				rtnVal = fsm1;
			}
			break;
		}
		case Empty: {
			rtnVal = 0;
			break;
		}
	}
	return rtnVal;;
}

/* Evaluate an or block item of a regular expression. */
Fsm *ReOrItem::walk( ParseData *pd )
{
	/* The return value, is the alphabet signed? */
	Fsm *rtnVal = 0;
	bool isSigned = pd->isAlphSigned();
	switch ( type ) {
	case Data: {
		int arr[str.getLen()];
		makeFsmKeyArray( arr, str, str.getLen(), pd );
		rtnVal = Fsm::orFsm( arr, str.getLen(), isSigned );
		break;
	}
	case Range: {
		int lowerKey = makeFsmKeyChar( lower, pd );
		int upperKey = makeFsmKeyChar( upper, pd );
		rtnVal = Fsm::rangeFsm( lowerKey, upperKey, isSigned );
		break;
	}}
	return rtnVal;
}



