/*
 *  Copyright 2001, 2002 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#ifndef _PARSETREE_H
#define _PARSETREE_H

#include "ragel.h"
#include "fsmwrite.h"
#include "astring.h"
#include "avlmap.h"
#include "bstmap.h"
#include "astring.h"

#define FUNC_NUM_CLEAR -1

/* Types of builtin machines. */
enum BuiltinType
{
	BT_Any,
	BT_Ascii,
	BT_Extend,
	BT_Alpha,
	BT_Digit,
	BT_Alnum,
	BT_Lower,
	BT_Upper,
	BT_Cntrl,
	BT_Graph,
	BT_Print,
	BT_Punct,
	BT_Space,
	BT_Xdigit
};

/* Element in list of actions. Contains the string for the 
 * code to exectute. */
struct StringListEl : public DListEl<StringListEl>
{
public:
	StringListEl( char *data ) : data(data) { }
	~StringListEl() { delete[] data; }

	char *data;
};

/* List of string elements. Used for functions and data to go
 * into the fsm struct. */
typedef DList<StringListEl> StringList;

/* Global counting the function ordering. */
extern int curFuncOrdering;

/* Type of augmentation. Describes locations in the machine. */
enum AugType
{
	start,
	fin,
	all,
	leave,
	clearLeave
};

/* Store the value and type of a priority augmentation. */
struct PriorityAug
{
	PriorityAug( int val, AugType type ) :
		value(val), type(type) { }

	int value;
	AugType type;
};

/* Structrue represents a function assigned to some FactorWithAug
 * node. The factor with aug will keep an array of these. */
struct ParserFunc
{
	ParserFunc( int func, int ordering, AugType type ) :
		func(func), ordering(0), type(type) { }

	int func;
	int ordering;
	AugType type;
};

struct ExpressionNode;
struct TermNode;
struct FactorWithAugNode;
struct FactorWithRepNode;
struct FactorNode;
struct Literal;
struct Range;
struct RegExp;
struct ReItem;
struct ReOrBlock;
struct ReOrItem;

typedef AvlMapNode<char*, ExpressionNode*> GraphDictNode;
typedef AvlMap<char*, ExpressionNode*, CmpStr> GraphDict;

/*
 * FuncDict
 */
typedef BstMapEl<char*, int> FuncDictEl;
typedef BstMap<char*, int, CmpStr> FuncDict;

enum AlphType
{
	AT_Char,
	AT_UnsignedChar,
	AT_Short,
	AT_UnsignedShort,
	AT_Int,
	AT_UnsignedInt
};

/* Class to collect information about the machine during the 
 * parse of input. */
class ParseData
{
public:
	/* Create a new parse data object. This is done at the beginning of every
	 * fsm specification. */
	ParseData( const String &fileName );
	~ParseData();

	/*
	 * Setting up the graph dict.
	 */

	/* Initialize a graph dict with the basic fsms. */
	void initGraphDict( );

	/* Set the alphabet type. If type types are not valid returns false. */
	bool setAlphType( String s1, String s2 );
	bool setAlphType( String s1 );

	/*
	 * Generate an fsm and writing it out.
	 */

	/* Generate and write out the fsm. */
	void generateFsm( );

	/* Set the lower and upper range from the lower and upper number keys. */
	void setLowerUpperRange( );

	/* dump a graph to std out. */
	void dumpWithHeader( FsmAp *graph, const char *head );

	/*
	 * Querying the parse data
	 */

	/* Is the alphabet a signed type? */
	bool isAlphSigned();
	
	/*
	 * Data collected during the parse.
	 */

	/* Dictionary of graphs. When done we will lookup "main" */
	GraphDict graphDict;

	/* Dictionary of functions. Lets functions be defined and then 'called'.
	 * They are not really called though, just used. */
	FuncDict funcDict;

	/* List of functions. Will be pasted into a switch statement. */
	StringList funcList;

	/* List of sections of data to add to the fsm struct. */
	StringList dataList;

	/* The func id of the next function. */
	int nextFuncNum;

	/* If a machine is given this is set. */
	bool machineGiven;

	/* Code sections to put before and after the function execute section. */
	StringList initCodeList;

	/* Alphabet type. */
	AlphType alphType;

	/* The alphabet range. */
	String lowerNum, upperNum;
	int lowerKey, upperKey;
	int rangeLineNum;

	/* The name of the fsm. */
	String fsmName;

	/* The name of the file the fsm is from. */
	String fileName;
};

/*
 * ExpressionNode
 */
struct ExpressionNode
{
	enum ExpressionNodeType { Or, Intersect, Subtract, Term, Builtin };

	/* Construct with an expression on the left and a term on the right. */
	ExpressionNode( ExpressionNode *expression, TermNode *term, 
		ExpressionNodeType type) : expression(expression), 
		term(term), builtin(builtin), type(type) { }

	/* Construct with only a term. */
	ExpressionNode( TermNode *term ) :
		expression(0), term(term), builtin(builtin), type(Term) { }
	
	/* Construct with a builtin type. */
	ExpressionNode( BuiltinType builtin ) :
		expression(0), term(0), builtin(builtin), type(Builtin) { }

	~ExpressionNode();

	/* Evaluate the expression. */
	Fsm *walk( ParseData *pd );

	/* Data stored by the node. */
	ExpressionNode *expression;
	TermNode *term;
	BuiltinType builtin;
	ExpressionNodeType type;
};

/*
 * TermNode
 */
struct TermNode 
{
	enum TermNodeType { Concat, FactorWithAug };

	TermNode( bool leavingFsm, TermNode *term, FactorWithAugNode *factorWithAug ) :
		leavingFsm(leavingFsm), term(term), 
		factorWithAug(factorWithAug), type(Concat) { }

	TermNode( FactorWithAugNode *factorWithAug ) :
		leavingFsm(true), term(0), 
		factorWithAug(factorWithAug), type(FactorWithAug) { }
	
	~TermNode();

	Fsm *walk( ParseData *pd );

	bool leavingFsm;
	TermNode *term;
	FactorWithAugNode *factorWithAug;
	TermNodeType type;
};

/*
 * FactorWithAugNode
 */
struct FactorWithAugNode
{
	FactorWithAugNode(FactorWithRepNode *factorWithRep ) :
		funcs(), priorityAugs(), factorWithRep(factorWithRep) { }

	~FactorWithAugNode();

	Fsm *walk( ParseData *pd );

	Vector<ParserFunc> funcs;
	Vector<PriorityAug> priorityAugs;

	FactorWithRepNode *factorWithRep;
	FactorWithAugNode *factorWithAug;
};

/*
 * FactorWithRepNode
 */
struct FactorWithRepNode
{
	enum FactorWithRepType
		{ Star, Optional, Plus, Negate, Factor };

	FactorWithRepNode( bool leavingFsm, FactorWithRepNode *factorWithRep,
			FactorWithRepType type ) :
		leavingFsm(leavingFsm), factorWithRep(factorWithRep), 
		factor(0), type(type) { }

	FactorWithRepNode( FactorNode *factor ) :
		leavingFsm(true), factorWithRep(0), factor(factor), type(Factor) { }

	~FactorWithRepNode();

	Fsm *walk( ParseData *pd );

	bool leavingFsm;
	FactorWithRepNode *factorWithRep;
	FactorNode *factor;
	FactorWithRepType type;
};

/*
 * FactorNode
 */
struct FactorNode 
{
	/* Language elements a factor node can be. */
	enum FactorType {
		LiteralFsm, 
		RangeFsm, 
		RegularExpression, 
		Expression, 
		LookupExpression
	}; 

	FactorNode( RegExp *regExp ) :
		regExp(regExp), type(RegularExpression) { }

	/* Construct with a literal fsm. */
	FactorNode( Literal *literal ) :
		literal(literal), type(LiteralFsm) { }

	/* Construct with a range. */
	FactorNode( Range *range ) : 
		range(range), type(RangeFsm) { }

	/* Construct with an expression. */
	FactorNode( ExpressionNode *expression, FactorType type ) :
		expression(expression), type(type) {}

	/* Cleanup. */
	~FactorNode();

	/* Evaluation. */
	Fsm *walk( ParseData *pd );

	Literal *literal;
	Range *range;
	RegExp *regExp;
	ExpressionNode *expression;
	int lower, upper;
	FactorType type;
};

/* A range machine. Only ever composed of two literals. */
struct Range
{
	Range( Literal *lowerLit, Literal *upperLit ) 
		: lowerLit(lowerLit), upperLit(upperLit) { }

	~Range();
	Fsm *walk( ParseData *pd );

	Literal *lowerLit;
	Literal *upperLit;
};

/* Some literal machine. Can be a number or literal string. */
struct Literal
{
	enum LiteralType { Number, LitString, OrLitString };

	Literal( int lineNum, String str, LiteralType type )
		: str(str), type(type) { }

	int makeRangeEndPoint( ParseData *pd );
	Fsm *walk( ParseData *pd );
	
	int lineNum;
	String str;
	LiteralType type;
};

/* Regular expression. */
struct RegExp
{
	enum RegExpType { RecurseItem, Empty };

	/* Constructors. */
	RegExp() 
		: type(Empty) { }
	RegExp(RegExp *regExp, ReItem *item) 
		: regExp(regExp), item(item), type(RecurseItem) { }

	~RegExp();
	Fsm *walk( ParseData *pd );

	RegExp *regExp;
	ReItem *item;
	RegExpType type;
};

/* An item in a regular expression. */
struct ReItem
{
	enum ReItemType { Data, Dot, OrBlock, NegOrBlock };
	
	ReItem( char c ) 
		: str(c), star(false), type(Data) { }
	ReItem( ReItemType type )
		: star(false), type(type) { }
	ReItem( ReOrBlock *orBlock, ReItemType type )
		: orBlock(orBlock), star(false), type(type) { }

	~ReItem();
	Fsm *walk( ParseData *pd );

	String str;
	ReOrBlock *orBlock;
	bool star;
	ReItemType type;
};

/* An or block item. */
struct ReOrBlock
{
	enum ReOrBlockType { RecurseItem, Empty };

	/* Constructors. */
	ReOrBlock()
		: type(Empty) { }
	ReOrBlock(ReOrBlock *orBlock, ReOrItem *item)
		: orBlock(orBlock), item(item), type(RecurseItem) { }

	~ReOrBlock();
	Fsm *walk( ParseData *pd );
	
	ReOrBlock *orBlock;
	ReOrItem *item;
	ReOrBlockType type;
};

/* An item in an or block. */
struct ReOrItem
{
	enum ReOrItemType { Data, Range };

	ReOrItem( char c ) 
		: str(c), type(Data) { }
	ReOrItem( char lower, char upper )
		: lower(lower), upper(upper), type(Range) { }

	Fsm *walk( ParseData *pd );

	String str;
	char lower;
	char upper;
	ReOrItemType type;
};




#endif /* _PARSETREE_H */
