/*
 *  Copyright 2002 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Reglang FSM.
 *
 *  Reglang FSM is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Reglang FSM is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Reglang FSM; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#ifndef _RLFSM_FSMMIN_CPP
#define _RLFSM_FSMMIN_CPP

template < class Graph, class State, class Transition >
		int FsmGraph<Graph, State, Transition>::
		partitionRound( State **statePtrs, MinPartition<State> *parts, int numParts )
{
	/* Need a mergesort object and a single partition compare. */
	MergeSort< State*, PartitionCompare<State, Transition> > mergeSort;
	PartitionCompare<State, Transition> partCompare;

	/* Set the signed/unsigned flags of the mergesort and compare. */
	mergeSort.isAlphSigned = isAlphSigned;
	partCompare.isAlphSigned = isAlphSigned;

	/* For each partition. */
	for ( int p = 0; p < numParts; p++ ) {
		/* Fill the pointer array with the states in the partition. */
		int s, numStates = parts[p].list.length;
		State *state = parts[p].list.head;
		for ( s = 0; state != 0; s++, state = state->next )
			statePtrs[s] = state;

		/* Sort the states using the partitioning compare. */
		mergeSort.sort( statePtrs, numStates );

		/* Assign the states into partitions based on the results of the sort. */
		int destPart = p, firstNewPart = numParts;
		for ( s = 1; s < numStates; s++ ) {
			/* If this state differs from the last then move to the next partition. */
			if ( partCompare.compare( statePtrs[s-1], statePtrs[s] ) < 0 ) {
				/* The new partition is the next avail spot. */
				destPart = numParts;
				numParts += 1;
			}

			/* If the state is not staying in the first partition, then
			 * transfer it to its destination partition. */
			if ( destPart != p ) {
				State *state = parts[p].list.detach( statePtrs[s] );
				parts[destPart].list.append( state );
			}
		}

		/* Fix the partition pointer for all the states that got moved to a new
		 * partition. This must be done after the states are transfered so the
		 * result of the sort is not altered. */
		for ( int newPart = firstNewPart; newPart < numParts; newPart++ ) {
			State *state = parts[newPart].list.head;
			while ( state != 0 ) {
				state->alg.partition = &parts[newPart];
				state = state->next;
			}
		}
	}

	return numParts;
}

template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		minimizePartition1()
{
	/* Need one mergesort object and partition compares. */
	MergeSort< State*, InitPartitionCompare<State, Transition> > mergeSort;
	InitPartitionCompare<State, Transition> initPartCompare;

	/* Nothing to do if there are no states. */
	if (stateList.length == 0)
		return;

	/* Set the signed/unsigned flag of the mergesort and compare. */
	mergeSort.isAlphSigned = isAlphSigned;
	initPartCompare.isAlphSigned = isAlphSigned;

	/* 
	 * First thing is to partition the states by final state status and
	 * transition functions. This gives us an initial partitioning to work
	 * with.
	 */

	/* Fill up an array of pointers to the states for easy sorting. */
	int s, numStates = stateList.length;
	State** statePtrs = new State*[numStates];
	State *state = stateList.head;
	for ( s = 0; s < numStates; s++ ) {
		statePtrs[s] = state;
		state = state->next;
	}
		
	/* Sort the states using the array of states. */
	mergeSort.sort( statePtrs, numStates );

	/* An array of lists of states is used to partition the states. */
	MinPartition<State> *parts = new MinPartition<State>[numStates];

	/* Assign the states into partitions. */
	int destPart = 0;
	for ( s = 0; s < numStates; s++ ) {
		/* If this state differs from the last then move to the next partition. */
		if ( s > 0 && initPartCompare.compare( statePtrs[s-1], statePtrs[s] ) < 0 ) {
			/* Move to the next partition. */
			destPart += 1;
		}

		/* Put the state into its partition. */
		statePtrs[s]->alg.partition = &parts[destPart];
		parts[destPart].list.append( statePtrs[s] );
	}

	/* We just moved all the states from the main list into partitions without
	 * taking them off the main list. So clean up the main list now. */
	stateList.abandon();

	/* Split partitions. */
	int numParts = destPart + 1;
	while ( true ) {
		/* Test all partitions for splitting. */
		int newNum = partitionRound( statePtrs, parts, numParts );

		/* When no partitions can be split, stop. */
		if ( newNum == numParts )
			break;

		numParts = newNum;
	}

	/* Fuse states in the same partition. The states will end up back on the
	 * main list. */
	fusePartitions( parts, numParts );

	/* Cleanup. */
	delete[] statePtrs;
	delete[] parts;
}

/* Split partitions that need splittting, decide which partitions might need
 * to be split as a result, continue until there are no more that might need
 * to be split. */
template < class Graph, class State, class Transition >
		int FsmGraph<Graph, State, Transition>::
		splitCandidates( State **statePtrs, MinPartition<State> *parts, int numParts )
{
	/* Need a mergesort and a partition compare. */
	MergeSort< State*, PartitionCompare<State, Transition> > mergeSort;
	PartitionCompare<State, Transition> partCompare;

	/* Set the signed/unsigned flag of the mergesort and compare. */
	mergeSort.isAlphSigned = isAlphSigned;
	partCompare.isAlphSigned = isAlphSigned;

	/* The lists of unsplitable (partList) and splitable partitions. 
	 * Only partitions in the splitable list are check for needing splitting. */
	DList< MinPartition<State> > partList, splittable;

	/* Initially, all partitions are born from a split (the initial
	 * partitioning) and can cause other partitions to be split. So any
	 * partition with a state with a transition out to another partition is a
	 * candidate for splitting. This will make every partition except possibly
	 * partitions of final states split candidates. */
	for ( int p = 0; p < numParts; p++ ) {
		/* Assume not active. */
		parts[p].active = false;

		/* Look for a trans out of any state in the partition. */
		State *state = parts[p].list.head;
		while ( state != 0 ) {
			/* If there is at least one transition out to another state then 
			 * the partition becomes splittable. */
			FsmOutIterator<State, Transition> outIt( state );
			if ( ! outIt.atEnd() ) {
				parts[p].active = true;
				break;
			}
			state = state->next;
		}

		/* If it was found active then it goes on the splittable list. */
		if ( parts[p].active )
			splittable.append( &parts[p] );
		else
			partList.append( &parts[p] );
	}

	/* While there are partitions that are splittable, pull one off and try
	 * to split it. If it splits, determine which partitions may now be split
	 * as a result of the newly split partition. */
	while ( splittable.length > 0 ) {
		MinPartition<State> *partition = splittable.detachFirst();

		/* Fill the pointer array with the states in the partition. */
		int s, numStates = partition->list.length;
		State *state = partition->list.head;
		for ( s = 0; state != 0; s++, state = state->next )
			statePtrs[s] = state;

		/* Sort the states using the partitioning compare. */
		mergeSort.sort( statePtrs, numStates );

		/* Assign the states into partitions based on the results of the sort. */
		MinPartition<State> *destPart = partition;
		int firstNewPart = numParts;
		for ( s = 1; s < numStates; s++ ) {
			/* If this state differs from the last then move to the next partition. */
			if ( partCompare.compare( statePtrs[s-1], statePtrs[s] ) < 0 ) {
				/* The new partition is the next avail spot. */
				destPart = &parts[numParts];
				numParts += 1;
			}

			/* If the state is not staying in the first partition, then
			 * transfer it to its destination partition. */
			if ( destPart != partition ) {
				State *state = partition->list.detach( statePtrs[s] );
				destPart->list.append( state );
			}
		}

		/* Fix the partition pointer for all the states that got moved to a new
		 * partition. This must be done after the states are transfered so the
		 * result of the sort is not altered. */
		int newPart;
		for ( newPart = firstNewPart; newPart < numParts; newPart++ ) {
			State *state = parts[newPart].list.head;
			while ( state != 0 ) {
				state->alg.partition = &parts[newPart];
				state = state->next;
			}
		}

		/* Put the partition we just split and any new partitions that came out
		 * of the split onto the inactive list. */
		partition->active = false;
		partList.append( partition );
		for ( newPart = firstNewPart; newPart < numParts; newPart++ ) {
			parts[newPart].active = false;
			partList.append( &parts[newPart] );
		}

		if ( destPart == partition )
			continue;

		/* Now determine which partitions are splittable as a result of
		 * splitting partition by walking the in lists of the states in
		 * partitions that got split. Partition is the faked first item in the
		 * loop. */
		MinPartition<State> *causalPart = partition;
		newPart = firstNewPart - 1;
		while ( newPart < numParts ) {
			State *state = causalPart->list.head;
			while ( state != 0 ) {
				FsmInIterator<State, Transition> inIt( state );
				for ( ; !inIt.atEnd(); inIt++ ) {
					MinPartition<State> *fromPart = 
							inIt.trans->fromState->alg.partition;
					if ( ! fromPart->active ) {
						fromPart->active = true;
						partList.detach( fromPart );
						splittable.append( fromPart );
					}
				}
				state = state->next;
			}

			newPart += 1;
			causalPart = &parts[newPart];
		}
	}
	return numParts;
}


template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		minimizePartition2()
{
	/* Need a mergesort and an initial partition compare. */
	MergeSort< State*, InitPartitionCompare<State, Transition> > mergeSort;
	InitPartitionCompare<State, Transition> initPartCompare;

	/* Set the signed/unsigned flag of the mergesort and compare. */
	mergeSort.isAlphSigned = isAlphSigned;
	initPartCompare.isAlphSigned = isAlphSigned;

	/* Nothing to do if there are no states. */
	if (stateList.length == 0)
		return;

	/* 
	 * First thing is to partition the states by final state status and
	 * transition functions. This gives us an initial partitioning to work
	 * with.
	 */

	/* Fill up an array of pointers to the states for easy sorting. */
	int s, numStates = stateList.length;
	State** statePtrs = new State*[numStates];
	State *state = stateList.head;
	for ( s = 0; s < numStates; s++ ) {
		statePtrs[s] = state;
		state = state->next;
	}
		
	/* Sort the states using the array of states. */
	mergeSort.sort( statePtrs, numStates );

	/* An array of lists of states is used to partition the states. */
	MinPartition<State> *parts = new MinPartition<State>[numStates];

	/* Assign the states into partitions. */
	int destPart = 0;
	for ( s = 0; s < numStates; s++ ) {
		/* If this state differs from the last then move to the next partition. */
		if ( s > 0 && initPartCompare.compare( statePtrs[s-1], statePtrs[s] ) < 0 ) {
			/* Move to the next partition. */
			destPart += 1;
		}

		/* Put the state into its partition. */
		statePtrs[s]->alg.partition = &parts[destPart];
		parts[destPart].list.append( statePtrs[s] );
	}

	/* We just moved all the states from the main list into partitions without
	 * taking them off the main list. So clean up the main list now. */
	stateList.abandon();

	/* Split partitions. */
	int numParts = splitCandidates( statePtrs, parts, destPart+1 );

	/* Fuse states in the same partition. The states will end up back on the
	 * main list. */
	fusePartitions( parts, numParts );

	/* Cleanup. */
	delete[] statePtrs;
	delete[] parts;
}

template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		initialMarkRound( MarkIndex<State> &markIndex )
{
	/* P and q for walking pairs. */
	State *p = stateList.head, *q;

	/* Need an initial partition compare. Set the signed/unsigned flag. */
	InitPartitionCompare<State, Transition> initPartCompare;
	initPartCompare.isAlphSigned = isAlphSigned;

	/* Walk all unordered pairs of (p, q) where p != q.
	 * The second depth of the walk stops before reaching p. This
	 * gives us all unordered pairs of states (p, q) where p != q. */
	while ( p != 0 ) {
		q = stateList.head;
		while ( q != p ) {
			/* If the states differ on final state status, out transitions or
			 * any transition data then they should be separated on the initial
			 * round. */
			if ( initPartCompare.compare( p, q ) != 0 )
				markIndex.markPair( p->alg.stateNum, q->alg.stateNum );

			q = q->next;
		}
		p = p->next;
	}
}

template < class Graph, class State, class Transition >
		bool FsmGraph<Graph, State, Transition>::
		markRound( MarkIndex<State> &markIndex )
{
	/* P an q for walking pairs. Take note if any pair gets marked. */
	State *p = stateList.head, *q;
	bool pairWasMarked = false;

	/* Need a mark comparison. Set the signed/unsigned flag. */
	MarkCompare<State, Transition> markCompare;
	markCompare.isAlphSigned = isAlphSigned;

	/* Walk all unordered pairs of (p, q) where p != q.
	 * The second depth of the walk stops before reaching p. This
	 * gives us all unordered pairs of states (p, q) where p != q. */
	while ( p != 0 ) {
		q = stateList.head;
		while ( q != p ) {
			/* Should we mark the pair? */
			if ( !markIndex.isPairMarked( p->alg.stateNum, q->alg.stateNum ) ) {
				if ( markCompare.shouldMark( markIndex, p, q ) ) {
					markIndex.markPair( p->alg.stateNum, q->alg.stateNum );
					pairWasMarked = true;
				}
			}
			q = q->next;
		}
		p = p->next;
	}

	return pairWasMarked;
}


template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		minimizeStable()
{
	/* Set the state numbers. */
	setStateNumbers();

	/* This keeps track of which pairs have been marked. */
	MarkIndex<State> markIndex( stateList.length );

	/* Mark pairs where final stateness, out trans, or trans data differ. */
	initialMarkRound( markIndex );

	/* While the last round of marking succeeded in marking a state
	 * continue to do another round. */
	int modified = markRound( markIndex );
	while (modified)
		modified = markRound( markIndex );

	/* Merge pairs that are unmarked. */
	fuseUnmarkedPairs( markIndex );
}

template < class Graph, class State, class Transition >
		bool FsmGraph<Graph, State, Transition>::
		minimizeRound()
{
	/* Nothing to do if there are no states. */
	if (stateList.length == 0)
		return false;

	/* Need a mergesort on approx compare and an approx compare. */
	MergeSort< State*, ApproxCompare<State, Transition> > mergeSort;
	ApproxCompare<State, Transition> approxCompare;

	/* Set the signed/unsigned flags of the mergesort and compare. */
	mergeSort.isAlphSigned = isAlphSigned;
	approxCompare.isAlphSigned = isAlphSigned;

	/* Fill up an array of pointers to the states. */
	State **statePtrs = new State*[stateList.length];
	State *state = stateList.head;
	State **dst = statePtrs;
	while ( state != 0 ) {
		*dst = state;
		dst += 1;
		state = state->next;
	}

	bool modified = false;

	/* Sort The list. */
	mergeSort.sort( statePtrs, stateList.length );

	/* Walk the list looking for duplicates next to each other, 
	 * merge in any duplicates. */
	State **pLast = statePtrs;
	State **pState = statePtrs + 1;
	for ( int i = 1; i < stateList.length; i++, pState++ ) {
		if ( approxCompare.compare( *pLast, *pState ) == 0 ) {
			/* Last and pState are the same, so fuse together. Move forward
			 * with pState but not with pLast. If any more are identical, we
			 * must */
			fuseEquivStates( *pLast, *pState );
			modified = true;
		}
		else {
			/* Last and this are different, do not set to merge them. Move
			 * pLast to the current (it may be way behind from merging many
			 * states) and pState forward one to consider the next pair. */
			pLast = pState;
		}
	}
	delete[] statePtrs;
	return modified;
}

template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		minimizeApproximate()
{
	/* While the last minimization round succeeded in compacting states,
	 * continue to try to compact states. */
	while ( true ) {
		bool modified = minimizeRound();
		if ( ! modified )
			break;
	}
}


/* Remove states that have no path to them from the start state. Recursively
 * traverses the graph marking states that have paths into them. Then removes
 * all states that did not get marked. */
template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		removeUnreachableStates()
{
	State *state, *next;

	/* Mark all the states that can be reached
	 * through the existing transition set. */
	markReachableFromHere( findEntry(0) );

	/* Delete all states that are not marked
	 * and unmark the ones that are marked. */
	state = stateList.head;
	while ( state ) {
		next = state->next;

		if ( state->stateBits & SB_ISMARKED )
			state->stateBits &= ~ SB_ISMARKED;
		else {
			detachState( state );
			stateList.detach( state );
			delete state;
		}

		state = next;
	}
}

/* Remove states that that do not lead to a final states. Works recursivly traversing
 * the graph in reverse (starting from all final states) and marking seen states. Then
 * removes states that did not get marked. */
template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		removeDeadEndStates()
{
	State *state, *next;

	/* Mark all states that have paths to the final states. */
	State **st = finStateSet.data;
	int nst = finStateSet.length;
	for ( int i = 0; i < nst; i++, st++ )
		markReachableFromHereReverse( *st );

	/* Start state gets honorary marking. If the machine accepts nothing we
	 * still want the start state to hang around. This must be done before the
	 * recursive call on all the final states so that it does not cause the
	 * start state in transitions to be skipped when the start state is
	 * visited by the traversal. */
	findEntry(0)->stateBits |= SB_ISMARKED;

	/* Delete all states that are not marked
	 * and unmark the ones that are marked. */
	state = stateList.head;
	while (state) {
		next = state->next;

		if ( state->stateBits & SB_ISMARKED  )
			state->stateBits &= ~ SB_ISMARKED;
		else {
			detachState( state );
			stateList.detach( state );
			delete state;
		}
		
		state = next;
	}
}

/* Remove states on the misfit list. To work properly misfit accounting should
 * be on when this is called. The detaching of a state will likely cause
 * another misfit to be collected and it can then be removed. */
template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		removeMisfits()
{
	while ( misfitList.length > 0 ) {
		/* Get the first state. */
		State *state = misfitList.head;

		/* Detach and delete. */
		detachState( state );

		/* The state was previously on the misfit list and detaching can only
		 * remove in transitions so the state must still be on the misfit
		 * list. */
		misfitList.detach( state );
		delete state;
	}
}

/* Fuse src into dest because they have been deemed equivalent states.
 * Involves moving transitions into src to go into dest and invoking
 * callbacks. Src is deleted detached from the graph and deleted. */
template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		fuseEquivStates( State *dest, State *src )
{
	/* This would get ugly. */
	assert( dest != src );

	/* Cur is a duplicate. We can merge it with trail. */
	inTransMove( dest, src );

	/* For each entry point into src, move it to dest. */
	while ( src->entryIds.length > 0 ) {
		/* Change the first one. */
		int entry = src->entryIds.data[0];
		changeEntry( entry, dest );
	}

	/* Now that the ids are copied, clear src's entry id set. */
	src->entryIds.empty();
	
	/* Call user routine. */
	dest->fuseInState( src );

	detachState( src );
	stateList.detach( src );
	delete src;
}

template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		fuseUnmarkedPairs( MarkIndex<State> &markIndex )
{
	State *p = stateList.head, *nextP, *q;

	/* Definition: The primary state of an equivalence class is the first state
	 * encounterd that belongs to the equivalence class. All equivalence
	 * classes have primary state including equivalence classes with one state
	 * in it. */

	/* For each unmarked pair merge p into q and delete p. q is always the
	 * primary state of it's equivalence class. We wouldn't have landed on it
	 * here if it were not, because it would have been deleted.
	 *
	 * Proof that q is the primaray state of it's equivalence class: Assume q
	 * is not the primary state of it's equivalence class, then it would be
	 * merged into some state that came before it and thus p would be
	 * equivalent to that state. But q is the first state that p is equivalent
	 * to so we have a contradiction. */

	/* Walk all unordered pairs of (p, q) where p != q.
	 * The second depth of the walk stops before reaching p. This
	 * gives us all unordered pairs of states (p, q) where p != q. */
	while ( p != 0 ) {
		nextP = p->next;

		q = stateList.head;
		while ( q != p ) {
			/* If one of p or q is a final state then mark. */
			if ( ! markIndex.isPairMarked( p->alg.stateNum, q->alg.stateNum ) ) {
				fuseEquivStates( q, p );
				break;
			}
			q = q->next;
		}
		p = nextP;
	}
}

template < class Graph, class State, class Transition >
		void FsmGraph<Graph, State, Transition>::
		fusePartitions( MinPartition<State> *parts, int numParts )
{
	/* For each partition, fuse state 2, 3, ... into state 1. */
	for ( int p = 0; p < numParts; p++ ) {
		/* Assume that there will always be at least one state. */
		State *first = parts[p].list.head, *toFuse = first->next;

		/* Put the first state back onto the main state list. Don't bother
		 * removing it from the partition list first. */
		stateList.append( first );

		/* Fuse the rest of the state into the first. */
		while ( toFuse != 0 ) {
			/* Save the next. We will trash it before it is needed. */
			State *next = toFuse->next;

			/* Put the state to be fused in to the first back onto the main
			 * list before it is fuse.  the graph. The state needs to be on
			 * the main list for the detach from the graph to work.  Don't
			 * bother removing the state from the partition list first. We
			 * need not maintain it. */
			stateList.append( toFuse );

			/* Now fuse to the first. */
			fuseEquivStates( first, toFuse );

			/* Go to the next that we saved before trashing the next pointer. */
			toFuse = next;
		}

		/* We transfered the states from the partition list into the main list without
		 * removing the states from the partition list first. Clean it up. */
		parts[p].list.abandon();
	}
}

#endif /* _RLFSM_FSMMIN_CPP */
