/*
 *  Copyright 2001, 2002 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

%{

#include <iostream>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include "ragel.h"
#include "fsmwrite.h"
#include "parsetree.h"

using std::cerr;
using std::endl;

/* Generate a fsm from a graph dict. In main.cc. */
void generateFsm( char *fsmName, ParseData *parseData );

/* Ids to use for builtin functions. */
int builtinNextId = 1;

/* The parse data. The parser collects the things that it parses in
 * the data structures in here. */
ParseData *parseData;

/* Minimization after an operator. */
void AfterOpMinimize(Fsm *fsm);

%}

%union {
	/* General data types. */
	char *data;
	char chr;
	int integer;
	Literal *literal;

	/* Tree nodes. */
	TermNode *term;
	FactorWithRepNode *factorWithRep;
	FactorWithAugNode *factorWithAug;
	FactorNode *factor;
	ExpressionNode *expression;

	/* Priorities and actions. */
	AugType augType;
	int funcId;

	/* Regular expression items. */
	RegExpSet regExpSet;
	RegExp *regExp;
	ReItem *reItem;
	ReOrBlock *reOrBlock;
	ReOrItem *reOrItem;
}

%token <data> PP_Literal
%token <data> PP_Integer
%token <data> PP_Word
%token <data> PP_Comment
%token <data> PP_Character
%token <data> PP_Whitespace
%token <data> PP_Section

%type <data> NonFsmAny
%type <data> FsmName

%token <data> RL_UInt
%token <data> RL_PInt
%token <data> RL_NInt
%token <data> RL_Hex
%token <data> RL_Word
%token <data> RL_Literal
%token <data> RL_DoubleDot

%token <chr> RL_ReLiteralSlash
%token <chr> RL_ReLiteralChar
%token <chr> RL_ReLiteralOpen
%token <chr> RL_ReLiteralOpenNeg
%token <chr> RL_ReLiteralClose
%token <chr> RL_ReLiteralDot
%token <chr> RL_ReLiteralStar
%token <regExpSet> RL_ReLiteralSet

%token <data> RL_OrLiteral
%token <data> RL_Builtin
%token <data> RL_Data
%token <data> RL_Func
%token <data> RL_Init
%token <data> RL_Clear
%token <data> RL_AlphType
%token <data> RL_Range
%token <data> '*' '?' '+' '!' '(' ')' ';' '=' ':' '%' '$' '-' '|' '&' '.' '>'

%type <augType> AugType
%type <data> PriorityAug
%type <funcId> FunctionCall
%type <expression> Expression 
%type <term> Term
%type <factorWithAug> FactorWithAug 
%type <factorWithRep> FactorWithRep 
%type <factor> Factor
%type <literal> RangeLit
%type <data> AlphabetNum

%type <regExp> RegularExpr
%type <reItem> RegularExprItem
%type <reItem> RegularExprChar
%type <reOrBlock> RegularExprOrData
%type <reOrItem> RegularExprOrChar

%%

/* Input is any number of input sections. An empty file is accepted. */
input: Sections ;
Sections: 		
	Sections SectionItem { } |
	{ }
	;

/* A section item consists of either a single item from the input file (could
 * be a comment, string, identifier) which is passed through to the output
 * file as is or could be a fsm specification in which case it is parsed. */
SectionItem:
	NonFsmAny {
		// Write the data out to the output stream. */
		*outputStream << $1;
	} |
	FsmSpec
	;

/* The types of tokens that can appear outside of a fsm specifications
 * They are the alternate to a fsm specification in the SetionItem. */
NonFsmAny:
	PP_Literal |
	PP_Integer |
	PP_Word |
	PP_Comment |
	PP_Character |
	PP_Whitespace
	;


/* Fsm Specification. Fsms are deliminated by the %% tokens 
 * and require a name for the fsm immediately following the open %% 
 * deliminator. */
FsmSpec:
	PP_Section FsmName StatementList PP_Section {
		// Set the fsm's name. */
		parseData->fsmName = $2;

		// Process the fsm, then delete it.
		parseData->generateFsm( );
		delete parseData;
	}
	;

/* The Name of the fsm. Use this production to initialize the globals
 * used during the parse of the fsm. */
FsmName:
	RL_Word {
		// Take this opportunity to new up everything
		// necessary to process the fsm
		parseData = new ParseData( inputFile );
		$$ = $1;
	}
	;


/* A NonEmpty list of statements in a fsm. */
StatementList:
	StatementList Statement |
	Statement
	;

/* The differnt types of statements in a fsm spec. */
Statement:	
	RLAssignment |
	DataSpec |
	FuncSpec |
	InitSpec |
	AlphSpec |
	RangeSpec;

/* Allow the user to give some code to insert into the init function. */
InitSpec:
	RL_Init RL_Builtin {
	// Put the code on the init code list. 
		parseData->initCodeList.append( new StringListEl( $2 ) );
	} ;

/* Allow the user to give some declarations to put into the fsm structure. */
DataSpec:	
	RL_Data RL_Builtin ';' {
		// Put the spec for the data on the list. Note that we use
		// a func list el but it is data.
		parseData->dataList.append( new StringListEl( $2 ) );
	} ;

/* Allow the user to create a named fsm function that can be referenced when
 * building a machine. */
FuncSpec:	
	RL_Func RL_Word RL_Builtin {
		if ( parseData->funcDict.insert( $2,
						parseData->nextFuncNum ) ) {
			// Add the function to the list of functions.
			parseData->funcList.append( new StringListEl( $3 ) );

			// Used up a function number.
			parseData->nextFuncNum += 1;
		}
		else {
			cerr << inputFile << ":" << rlCurLine << ": func \"" << 
				$2 << "\" already defined" << endl;
			exit(1);
		}
	} ;

/* Specifies the data type of the input alphabet. One or two words 
 * followed by a semi-colon. */
AlphSpec:
	RL_AlphType RL_Word RL_Word ';' {
		if ( ! parseData->setAlphType( $2, $3 ) ) {
			cerr << inputFile << ":" << rlCurLine << ": \"" << $2 << 
				" " << $3 << "\" is not an allowed alphabet type" << endl;
			exit(1);
		}
	} |
	RL_AlphType RL_Word ';' {
		if ( ! parseData->setAlphType( $2 ) ) {
			cerr << inputFile << ":" << rlCurLine << ": \"" << $2 << 
				"\" is not an allowed alphabet type" << endl;
			exit(1);
		}
	}
	;

RangeSpec:
	RL_Range AlphabetNum AlphabetNum ';' {
		/* Save the upper and lower ends of the range and emit the line number. */
		parseData->lowerNum = $2;
		parseData->upperNum = $3;
		parseData->rangeLineNum = rlCurLine;
	};

/* An assignement statement. Makes an fsm and assigns it to a variable. */
RLAssignment:
	RL_Word '=' Expression ';' {
		if ( ! parseData->graphDict.insert($1, $3) ) {
			cerr << inputFile << ":" << rlCurLine << ": variable \"" << $1 <<
				"\" previously defined" << endl;
			exit(1);
		}
		parseData->machineGiven = true;
	}
	;

/* Top level production in the parse of a fsm. The lowest precedence
 * is the '|' (or), '&' (intersection), and '-' (subtraction) operators. */
Expression:
	Expression '|' Term {
		$$ = new ExpressionNode( $1, $3, ExpressionNode::Or );
	} |
	Expression '&' Term {
		$$ = new ExpressionNode( $1, $3, ExpressionNode::Intersect );
	} |
	Expression '-' Term {
		$$ = new ExpressionNode( $1, $3, ExpressionNode::Subtract );
	} |
	Term {
		$$ = new ExpressionNode( $1 );
	}
	;

/* The second group up in precedence. The concatenation operators. The special
 * '.^' operator forces the leaving fsm flag to the Concat operation to be 
 * false. The result is that the concatenation ignores out functions and out
 * priorities. */
Term:
	Term FactorWithAug {
		$$ = new TermNode( true, $1, $2 );
	} |
	Term '.' FactorWithAug {
		$$ = new TermNode( true, $1, $3 );
	} |
	Term '.' '^' FactorWithAug {
		$$ = new TermNode( false, $1, $4 );
	} |
	FactorWithAug {
		$$ = new TermNode( $1 );
	}
	;

/* Third group up in precenedce. Allow users to add function calls and to change
 * the priorities of transitions. */
FactorWithAug:
	FactorWithAug AugType FunctionCall {
		$$ = $1;
		$$->funcs.append( ParserFunc( $3, 0, $2 ) );
	} |
	FactorWithAug AugType PriorityAug {
		// Convert the priority number to a long. Check for overflow.
		errno = 0;
		int aug = strtol( $3, 0, 10 );
		if ( errno == ERANGE && aug == LONG_MAX ) {
			cerr << inputFile << ":" << rlCurLine << ": priority number " <<
					$3 << " overflows" << endl;
			exit(1);
		}
		if ( errno == ERANGE && aug == LONG_MIN ) {
			cerr << inputFile << ":" << rlCurLine << ": priority number " <<
					$3 << " underflows" << endl;
			exit(1);
		}
		
		// Append the priority augmentation to to the vector of priority augs
		// in the factor with aug and return the factor with aug.
		$1->priorityAugs.append( PriorityAug( aug, $2 ) );
		$$ = $1;
	} |
	FactorWithAug '%' '!' {
		$$ = $1;
		$$->priorityAugs.append( PriorityAug( 0, clearLeave ) );
	} |
	FactorWithRep {
		$$ = new FactorWithAugNode( $1 );
	}
	;

/* Priority change specs for FactorWithAug. */
PriorityAug: 
	RL_NInt | RL_PInt | RL_UInt ;
		
/* Differnt places to change the priorities or to add function calls. 
 * Used by FactorWithAug for both types of augmentation. */
AugType:
	':' { $$ = fin; } |
	'%' { $$ = leave; } |
	'$' { $$ = all; } |
	'>' { $$ = start; }
	;

/* Different types of function call specs. A RL_Word is reference to a func
 * given by the user as a statement in the fsm specification. A function
 * can also be specified immediately. */
FunctionCall:
	RL_Word {
		// Set the name in the funcDict.
		FuncDictEl *fdel = parseData->funcDict.find( $1 );
		if ( fdel != 0 ) {
			// Pass up the function id;
			$$ = fdel->value;
		}
		else {
			cerr << inputFile << ":" << rlCurLine << ": action lookup of \"" <<
					$1 << "\" failed" << endl;
			exit(1);
		}
	} |
	RL_Builtin {
		// Add the function to the list of functions.
		parseData->funcList.append( new StringListEl( $1 ) );

		// Pass up the builtin name.
		$$ = parseData->nextFuncNum;
		parseData->nextFuncNum += 1;
	} |
	RL_Clear {
		// This is a clear action, Don't set the action, clear
		// the actions instead.
		$$ = FUNC_NUM_CLEAR;
	}
	;

/* The fourth level of precedence. These are the unary operators that allow
 * for repetition and negating. */
FactorWithRep:
	'*' FactorWithRep {
		$$ = new FactorWithRepNode( 
					true, $2, FactorWithRepNode::Star );
	} |
	'*' '^' FactorWithRep {
		$$ = new FactorWithRepNode( 
					false, $3, FactorWithRepNode::Star );
	} |
	'?' FactorWithRep {
		$$ = new FactorWithRepNode( 
					true, $2, FactorWithRepNode::Optional );
	} |
	'+' FactorWithRep {
		$$ = new FactorWithRepNode( 
					true, $2, FactorWithRepNode::Plus );
	} |
	'+' '^' FactorWithRep {
		$$ = new FactorWithRepNode( 
					false, $3, FactorWithRepNode::Plus );
	} |
	'!' FactorWithRep {
		$$ = new FactorWithRepNode(
					true, $2, FactorWithRepNode::Negate );
	} |
	Factor {
		$$ = new FactorWithRepNode( $1 );
	}
	;

/* The highest level in precedence. Atomic machines such as references to other
 * machines, literal machines, regular expressions or Expressions in side of
 * parenthesis. */
Factor:
	RL_Literal {
		// Create a new factor node going to a concat literal. */
		$$ = new FactorNode( new Literal( rlCurLine, $1, Literal::LitString ) );
	} |

	RL_OrLiteral {
		// Create a new factor node going to an or literal. */
		$$ = new FactorNode( new Literal( rlCurLine, $1, Literal::OrLitString ) );
	} |

	RL_ReLiteralSlash RegularExpr RL_ReLiteralSlash {
		// Create a new factor node going to a regular exp.
		$$ = new FactorNode( $2 );
	} |

	AlphabetNum {
		// Create a new factor node going to a literal number. */
		$$ = new FactorNode( new Literal( rlCurLine, $1, Literal::Number ) );
	} |

	RangeLit RL_DoubleDot RangeLit {
		// Create a new factor node going to a range. */
		$$ = new FactorNode( new Range( $1, $3 ) );
	} |

	RL_Word {
		/* Find the named graph. */
		GraphDictNode *gdNode = parseData->graphDict.find( $1 );
		if ( gdNode == 0 ) {
			cerr << inputFile << ":" << rlCurLine << ": graph lookup of \"" << 
				$1 << "\" failed" << endl;
			exit(1);
		}

		/* Create a factor node that is a lookup of an expression. */
		$$ = new FactorNode( gdNode->value, FactorNode::LookupExpression );
	} |

	'(' Expression ')' {
		$$ = new FactorNode( $2, FactorNode::Expression );
	}
	;

/* Parser for regular expression fsms. Any number of expression items which
 * generally gives a machine one character long or one character long stared. */
RegularExpr:
	RegularExpr RegularExprItem {
		// An optimization to lessen the tree size. If a non-starred char is directly
		// under the left side on the right and the right side is another non-starred
		// char then paste them together and return the left side. Otherwise
		// just put the two under a new reg exp node.
		if ( $2->type == ReItem::Data && !$2->star &&
			$1->type == RegExp::RecurseItem &&
			$1->item->type == ReItem::Data && !$1->item->star )
		{
			// Append the right side to the right side of the left and toss 
			// the right side.
			$1->item->str += $2->str;
			delete $2;
			$$ = $1;
		}
		else {
			$$ = new RegExp( $1, $2 );
		}
	} |
	/* Nothing */ {
		// Can't optimize the tree.
		$$ = new RegExp();
	}
	;

/* RegularExprItems can be a character spec with an optional staring of the char. */
RegularExprItem:
	RegularExprChar RL_ReLiteralStar {
		$1->star = true;
		$$ = $1;
	} |
	RegularExprChar {
		$$ = $1;
	}
	;

/* A character spec can be a set of characters inside of square parenthesis,
 * a dot specifying any character or some explicitly stated character. */
RegularExprChar:
	RL_ReLiteralOpen RegularExprOrData RL_ReLiteralClose {
		$$ = new ReItem( $2, ReItem::OrBlock );
	} |
	RL_ReLiteralOpenNeg RegularExprOrData RL_ReLiteralClose {
		$$ = new ReItem( $2, ReItem::NegOrBlock );
	} |
	RL_ReLiteralDot {
		$$ = new ReItem( ReItem::Dot );
	} |
	RL_ReLiteralChar {
		$$ = new ReItem( $1 );
	}
	;

/* The data inside of a [] expression in a regular expression. Accepts any
 * number of characters or ranges. */
RegularExprOrData:
	RegularExprOrData RegularExprOrChar {
		// An optimization to lessen the tree size. If an or char is directly
		// under the left side on the right and the right side is another or
		// char then paste them together and return the left side. Otherwise
		// just put the two under a new or data node.
		if ( $2->type == ReOrItem::Data &&
				$1->type == ReOrBlock::RecurseItem &&
				$1->item->type == ReOrItem::Data )
		{
			// Append the right side to right side of the left and toss
			// the right side.
			$1->item->str += $2->str;
			delete $2;
			$$ = $1;
		}
		else {
			// Can't optimize, put the left and right under a new node.
			$$ = new ReOrBlock( $1, $2 );
		}
	} | 
	/* Nothing */ {
		$$ = new ReOrBlock();
	}
	;


/* A single character inside of an or expression. Can either be a character
 * or a set of characters. */
RegularExprOrChar:
	RL_ReLiteralChar {
		$$ = new ReOrItem( $1 );
	} |
	RL_ReLiteralSet {
		$$ = new ReOrItem( $1.lower, $1.upper );
	}
	;

RangeLit:
	RL_Literal {
		if ( strlen($1) != 1 ) {
			cerr << inputFile << ":" << rlCurLine << ": literal used in range "
					"must be of length 1" << endl;
			exit(1);
		}
		$$ = new Literal( rlCurLine, $1, Literal::LitString );
	} |

	AlphabetNum {
		// Create a new literal number. */
		$$ = new Literal( rlCurLine, $1, Literal::Number );
	}
	;

AlphabetNum:
	RL_UInt | RL_PInt | RL_NInt | RL_Hex ;
	
%%

void rlerror(char *err)
{
	cerr << inputFile << ":" << rlCurLine << ": " << err << endl;
	exit(1);
}
